/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import java.util.Collection;
import java.util.regex.Pattern;

import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.mime.MultipartEntityBuilder;

import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
class HttpPostRequestBuilder {

	static final String FORMCYCLE_ENDPOINT = "formData";
	static final String FORM_DATA_KEY = "formData";
	static final ContentType PROTOBUF_CONTENT_TYPE = ContentType.create("application/protobuf");
	static final String ATTACHMENT_KEY = "attachments";
	static final String REPRESENTATION_KEY = "representations";
	static final Pattern URL_WITH_ENDPOINT_PATTERN = Pattern.compile("^((https?://)?[^/]+)(/" + FORMCYCLE_ENDPOINT + ")$");
	private final MultipartEntityBuilder entityBuilder = MultipartEntityBuilder.create();
	private final OzgCloudConfig cloudConfig;

	public HttpPostRequestBuilder withFormData(FormCycleFormData formData) {
		entityBuilder.addBinaryBody(FORM_DATA_KEY, formData.toByteArray(), PROTOBUF_CONTENT_TYPE, null);
		return this;
	}

	public HttpPostRequestBuilder withAttachments(Collection<FormcycleAttachment> attachments) {
		for (FormcycleAttachment attachment : attachments) {
			entityBuilder.addBinaryBody(ATTACHMENT_KEY, attachment.getContent(), attachment.getContentType(), attachment.getFileName());
		}
		return this;
	}

	public HttpPostRequestBuilder withRepresentations(Collection<FormcycleAttachment> representations) {
		for (FormcycleAttachment representation : representations) {
			entityBuilder.addBinaryBody(REPRESENTATION_KEY, representation.getContent(), representation.getContentType(),
					representation.getFileName());
		}
		return this;
	}

	public HttpPost build() {
		var httpPost = new HttpPost(buildEingagsAdapterUrl());
		httpPost.setEntity(entityBuilder.build());
		return httpPost;
	}

	String buildEingagsAdapterUrl() {
		String eingangsAdapterUrl = cloudConfig.getEingangsAdapterUrl();
		if (URL_WITH_ENDPOINT_PATTERN.matcher(eingangsAdapterUrl).matches()) {
			return eingangsAdapterUrl;
		}
		return eingangsAdapterUrl.endsWith("/")
				? eingangsAdapterUrl + FORMCYCLE_ENDPOINT
				: eingangsAdapterUrl + "/" + FORMCYCLE_ENDPOINT;
	}
}
