/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static org.apache.commons.lang3.StringUtils.*;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Locale;
import java.util.Optional;
import java.util.function.Supplier;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.errorhandling.EWorkflowElementNodeError;
import de.ozgcloud.formcycle.errorhandling.EWorkflowElementNodeSoftError;
import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.errorhandling.UserErrorDispatcher;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.beans.interfaces.IGuiIcon;
import de.xima.fc.entities.Attachment;
import de.xima.fc.exceptions.AbstractAbruptCompletionException;
import de.xima.fc.exceptions.NodeThrewException;
import de.xima.fc.interfaces.plugin.lifecycle.IPluginInitializeData;
import de.xima.fc.interfaces.workflow.elements.IElementHelpLocation;
import de.xima.fc.interfaces.workflow.execution.IFileValueDescriptor;
import de.xima.fc.interfaces.workflow.execution.INormalCompletionResult;
import de.xima.fc.interfaces.workflow.execution.IWorkflowExecutionEnvironmentData;
import de.xima.fc.interfaces.workflow.mixin.IElementCategory;
import de.xima.fc.interfaces.workflow.mixin.IKeyValueSummarizableNode;
import de.xima.fc.interfaces.workflow.mixin.ISummaryKeyValueModel;
import de.xima.fc.interfaces.workflow.params.IGetElementPrototypesParams;
import de.xima.fc.interfaces.workflow.params.IGetElementSummaryParams;
import de.xima.fc.interfaces.workflow.params.INodeExecutionParams;
import de.xima.fc.interfaces.workflow.value.IUnionValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueBuilder;
import de.xima.fc.interfaces.workflow.value.IValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptorFactory;
import de.xima.fc.mdl.ui.GuiIcon;
import de.xima.fc.workflow.ElementHelpLocation;
import de.xima.fc.workflow.FileValueDescriptor;
import de.xima.fc.workflow.SummaryKeyValueModelBuilder;
import de.xima.fc.workflow.enums.EWorkflowElementCategory;
import de.xima.fc.workflow.mixin.INodePropertyPluginBean;
import de.xima.fc.workflow.mixin.IPluginActionNodeHandler;

public final class WorkflowElementNodePlugin
		implements IPluginActionNodeHandler<WorkflowElementNodeProps>, IKeyValueSummarizableNode<WorkflowElementNodeProps> {

	private static final Logger LOG = LoggerFactory.getLogger(WorkflowElementNodePlugin.class);

	private transient IPluginInitializeData initializeData;
	private transient PluginPropertiesMapper pluginPropertiesMapper;

	@Override
	public void initialize(IPluginInitializeData initializeData) {
		this.initializeData = initializeData;
		this.pluginPropertiesMapper = createPropertiesMapper();
	}

	PluginPropertiesMapper createPropertiesMapper() {
		return new PluginPropertiesMapper(new UserErrorDispatcher());
	}

	@Override
	public String getPropertiesViewXhtmlName() {
		return "WorkflowElementNode.xhtml";
	}

	@Override
	public String getName() {
		return "de.ozgcloud.formcycle.WorkflowElementNodePlugin";
	}

	@Override
	public IElementCategory getMainCategory(IGetElementPrototypesParams params) {
		// Main category where this node is shown in the drawer panel to the left of the
		// designer.
		return EWorkflowElementCategory.MAIN_ACTION;
	}

	@Override
	public IElementCategory getSubCategory(IGetElementPrototypesParams params) {
		// Sub category where this node is shown in the drawer panel to the left of the
		// designer.
		return EWorkflowElementCategory.SUB_ACTION_COMMON;
	}

	@Override
	public IElementHelpLocation getHelpPageLocation(Locale locale) throws IOException {
		// Optional HTML page documenting how to use this workflow node.
		// Use ElementHelpLocation#forExternalLink if you host the help page on an external server
		// Use ElementHelpLocation#forOfficialHelpPage when you host the help page on the FORMCYCLE wiki
		final var path = String.format("WEB-INF/properties/WorkflowElementNodeHelp_%s.html", locale.toLanguageTag());
		return ElementHelpLocation.forHtmlFile(getClass().getClassLoader(), path, StandardCharsets.UTF_8);
	}

	@Override
	public IGuiIcon getPrototypeIcon(IGetElementPrototypesParams params) {
		// Optional icon for this workflow node. Shown next to the node's name.
		return GuiIcon.valueOf("fa-question");
	}

	@Override
	public IFileValueDescriptor getFileValueDescriptor() {
		// Implement this method when this node can return files.
		// This node can then be selected by other actions that require files.
		// If this node does not provide files, remove this method
		return new FileValueDescriptor("WorkflowElementNodePlugin.files");
	}

	@Override
	public IValueDescriptor<?, ? extends IValueBuilder<?>> getSuccessValueDescriptor(IValueDescriptorFactory factory) {
		// Data returned by this node when it is executed
		// Users can access this data e.g. via placeholders, for example:
		// [%$Action.RESULT.name%]
		// [%$Action.RESULT.items.name%]
		// [%$Action.RESULT.items.state%]
		return factory.recordBuilder().requiredProperty(ExecutionResult.VORGANGNUMMER_PROPERTY_KEY, factory.string()).build();
	}

	@Override
	public IUnionValueDescriptor<String> getErrorValueDescriptor(IValueDescriptorFactory factory) {
		// For each error that can occur, additional data can be provided
		// Users can access this data e.g. via placeholders, for example:
		// [%$Action.ERROR_CODE%]
		// [%$Action.ERROR_MESSAGE%]
		// [%$Action.ERROR.file%]
		// [%$Action.ERROR.requiredPermissions[0]%]
		// [%$Action.ERROR.message%]
		// TODO das ist dafür gedacht, Fehler an der Oberfläche zu zeigen, was wir eigentlich nicht brauchen man könnte es einfach löschen
		return factory.unionStringBuilder() //
				.add(EWorkflowElementNodeError.OZGCLOUD_CANNOT_SEND.name(), f -> f.recordBuilder()
						.requiredProperty(NodeThrewExceptionFactory.ERROR_MESSAGE_KEY, f.string())
						.build()
				)
				.add(EWorkflowElementNodeError.PLUGIN_CANNOT_GET_ATTACHMENT.name(), f -> f.recordBuilder()
						.requiredProperty(NodeThrewExceptionFactory.ERROR_MESSAGE_KEY, f.string())
						.build()
				)
				.add(EWorkflowElementNodeError.PLUGIN_CONFIG_ERROR.name(), f -> f.recordBuilder()
						.requiredProperty(NodeThrewExceptionFactory.ERROR_MESSAGE_KEY, v -> v.string())
						.build())
				.build();
	}

	@Override
	public IUnionValueDescriptor<String> getSoftErrorValueDescriptor(IValueDescriptorFactory factory) {
		// For each error that can occur, additional data can be provided
		// Users can access this data e.g. via placeholders, for example:
		// [%$Action.ERROR_CODE%]
		// [%$Action.ERROR_MESSAGE%]
		// [%$Action.ERROR.file%]
		// [%$Action.ERROR.requiredPermissions[0]%]
		// [%$Action.ERROR.message%]
		return factory.unionStringBuilder() //
				.add(EWorkflowElementNodeSoftError.WRONG_HOST_NAME.name(), f -> f.recordBuilder() //
						.requiredProperty("expectedHostName", v -> v.string()) //
						.build()) //
				.add(EWorkflowElementNodeSoftError.USER_DOES_NOT_EXIST.name(), f -> f.string()) //
				.add(EWorkflowElementNodeError.PLUGIN_CONFIG_ERROR.name(), f -> f.recordBuilder()
						.requiredProperty(NodeThrewExceptionFactory.ERROR_MESSAGE_KEY, v -> v.string())
						.build())
				.build();
	}

	@Override
	public ISummaryKeyValueModel getElementSummaryKeyValueModel(IGetElementSummaryParams<WorkflowElementNodeProps> params) {
		// A simple summary table with the most important fields from the properties
		// This is shown when the user clicks on the info icon on a node in the designer.
		return new SummaryKeyValueModelBuilder().build();
	}

	@Override
	public INormalCompletionResult execute(INodeExecutionParams<WorkflowElementNodeProps> params) throws AbstractAbruptCompletionException {
		var workflowExecutionEnvironmentData = params.getWorkflowContext().env();
		var formProperties = params.getData();
		PluginFormDataAdapter pluginFormDataAdapter = new PluginFormDataAdapter(workflowExecutionEnvironmentData,
				formProperties.getOrganisationsEinheitId());
		var ozgCloudFormDataMapper = new OzgCloudFormDataMapper();
		var exceptionFactory = new NodeThrewExceptionFactory(params.throwingException());
		var attachmentsAdapter = new AttachmentMapper(exceptionFactory);
		var pluginExecutor = new WorkflowElementNodeExecutor(ozgCloudFormDataMapper, createOzgClient(formProperties, exceptionFactory),
				exceptionFactory,
				attachmentsAdapter, pluginFormDataAdapter, createAttachmentSupplier(workflowExecutionEnvironmentData));
		var executionResult = pluginExecutor.execute();
		return params.normalResult().success(executionResult.get()).build();
	}

	// TODO hier braucht man die Factory eigentlich nicht, es würde reichen einfaches Exception zu werfen
	OzgCloudFormDataHttpClient createOzgClient(WorkflowElementNodeProps formProperties, NodeThrewExceptionFactory exceptionFactory)
			throws NodeThrewException {
		var ozgCloudConfig = buildOzgCloudConfig(formProperties, exceptionFactory);
		return new OzgCloudFormDataHttpClient(ozgCloudConfig, new SystemPropertiesProvider());
	}

	OzgCloudConfig buildOzgCloudConfig(WorkflowElementNodeProps formProperties, NodeThrewExceptionFactory exceptionFactory)
			throws NodeThrewException {
		var eingangsAdapterUrl = getEingangsAdapterUrl(formProperties);
		if (isBlank(eingangsAdapterUrl)) {
			throw exceptionFactory.createConfigurationException("Eingangsadapter URL cannot be blank");
		}

		return OzgCloudConfig.builder()
				.eingangsAdapterUrl(eingangsAdapterUrl)
				.proxyConfig(pluginPropertiesMapper.mapProxyProperties(initializeData.getProperties()))
				.build();
	}

	String getEingangsAdapterUrl(WorkflowElementNodeProps formProperties) {
		if (formProperties.isUseCustomEingangsAdapterUrl()) {
			return formProperties.getEingangsAdapterUrl();
		}
		return pluginPropertiesMapper.getEingangsAdapterUrl(initializeData.getProperties());
	}

	private Supplier<List<Attachment>> createAttachmentSupplier(IWorkflowExecutionEnvironmentData environmentData) {
		return () -> Optional.ofNullable(environmentData.getFormRecord().getAttachments()).orElse(List.of());
	}

	@Override
	public Class<EWorkflowElementNodeError> getErrorCodeClass() {
		// Class of the enumeration with the possible types of errors that may occur.
		return EWorkflowElementNodeError.class;
	}

	@Override
	public Class<EWorkflowElementNodeSoftError> getSoftErrorCodeClass() {
		// Class of the enumeration with the possible types of soft errors that may occur.
		return EWorkflowElementNodeSoftError.class;
	}

	@Override
	public Class<? extends INodePropertyPluginBean<WorkflowElementNodeProps>> getMainPluginBeanClass() {
		// List of beans that should be available in the XHTML page.
		// Often you do not need a bean, in this case you can just remove this method.
		return WorkflowElementNodeBean.class;
	}

	@Override
	public IPluginInitializeData getPluginInitializeData() {
		return initializeData;
	}

}
