/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Supplier;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.attachment.AttachmentType;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.attachment.FormcycleAttachmentTestFactory;
import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.formdata.FormData;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.entities.Attachment;
import lombok.SneakyThrows;

class WorkflowElementNodeExecutorTest {

	@Spy
	@InjectMocks
	private WorkflowElementNodeExecutor executor;

	@Mock
	private OzgCloudFormDataMapper ozgCloudFormDataMapper;
	@Mock
	private OzgCloudFormDataHttpClient ozgHttpClient;
	@Mock
	private NodeThrewExceptionFactory exceptionFactory;
	@Mock
	private AttachmentMapper attachmentMapper;
	@Mock
	private PluginFormDataAdapter pluginFormDataAdapter;
	@Mock
	private Supplier<List<Attachment>> attachmentSupplier;
	@Captor
	private ArgumentCaptor<Map<String, Object>> successArgumentCaptor;

	@Nested
	class TestExecution {

		private static final String VORGANGS_NUMMER = "12345";

		@Mock
		private FormCycleConfirmationResponse response;
		@Mock
		private FormData formData;

		@Captor
		private ArgumentCaptor<List<FormcycleAttachment>> attachmentsCapture;

		@SneakyThrows
		@BeforeEach
		void setup() {
			when(response.getVorgangNummer()).thenReturn(VORGANGS_NUMMER);
			when(ozgHttpClient.send(any(), any(), any())).thenReturn(response);
			when(pluginFormDataAdapter.readFormData()).thenReturn(formData);
		}

		@Nested
		class TestMapping {

			@BeforeEach
			void setup() {
				when(attachmentSupplier.get()).thenReturn(List.of());
			}

			@SneakyThrows
			@Test
			void shouldExtractAsGrpc() {
				executor.execute();

				verify(ozgCloudFormDataMapper).map(formData);
			}

			@SneakyThrows
			@Test
			void shouldCreateSuccessResult() {
				var executionResult = executor.execute();

				assertThat(executionResult.get()).containsEntry(ExecutionResult.VORGANGNUMMER_PROPERTY_KEY, VORGANGS_NUMMER);
			}
		}

		@Nested
		@DisplayName("Send data")
		class TestSendData {

			@Mock
			private FormCycleFormData formcycleFormData;

			@Captor
			private ArgumentCaptor<Set<FormcycleAttachment>> attachmentsCapture;

			@SneakyThrows
			@BeforeEach
			void setup() {
				when(ozgCloudFormDataMapper.map(any())).thenReturn(formcycleFormData);
			}

			@SneakyThrows
			@Test
			void shouldSendFormData() {
				when(attachmentSupplier.get()).thenReturn(List.of());

				executor.execute();

				verify(ozgHttpClient).send(eq(formcycleFormData), any(), any());
			}

			@SneakyThrows
			@Test
			void shouldSendAttachments() {
				var expectedAttachment = FormcycleAttachmentTestFactory.create();
				doReturn(Map.of(
						AttachmentType.ATTACHMENT, Set.of(expectedAttachment),
						AttachmentType.REPRESENTATION, Set.of(FormcycleAttachmentTestFactory.create()))).when(executor).getAttachedFiles(any());

				executor.execute();

				verify(ozgHttpClient).send(any(), attachmentsCapture.capture(), any());
				assertThat(attachmentsCapture.getValue()).containsExactlyInAnyOrder(expectedAttachment);
			}

			@SneakyThrows
			@Test
			void shouldSendRepresentations() {
				var expectedRepresentations = FormcycleAttachmentTestFactory.create();
				doReturn(Map.of(
						AttachmentType.ATTACHMENT, Set.of(FormcycleAttachmentTestFactory.create()),
						AttachmentType.REPRESENTATION, Set.of(expectedRepresentations))).when(executor).getAttachedFiles(any());

				executor.execute();

				verify(ozgHttpClient).send(any(), any(), attachmentsCapture.capture());
				assertThat(attachmentsCapture.getValue()).containsExactlyInAnyOrder(expectedRepresentations);
			}

		}
	}

	@Nested
	@DisplayName("By error")
	class TestException {

		@Mock
		private FormData formData;

		@SneakyThrows
		@Test
		void shouldThrowException() {
			when(attachmentSupplier.get()).thenReturn(List.of());
			when(pluginFormDataAdapter.readFormData()).thenReturn(formData);
			when(ozgHttpClient.send(any(), any(), any())).thenThrow(IOException.class);

			try {
				executor.execute();
			} catch (NullPointerException e) {
			}

			verify(exceptionFactory).createCannotSendException(any(), any());
		}
	}

	@Nested
	@DisplayName("Attachment's handling")
	class TestGetAttachments {

		@Mock
		private Attachment attachment;
		@Mock
		private Attachment representation;

		@Test
		void shouldReturnEmpty() {
			when(attachmentSupplier.get()).thenReturn(Collections.emptyList());

			var attachedFiles = getAttachedFiles(Collections.emptySet());

			assertThat(attachedFiles).isEmpty();
		}

		@Test
		void shouldGetAttachments() {
			var expectedAttachment = initAttachment(attachment);

			var attachedFiles = getAttachedFiles(Set.of(FormcycleAttachmentTestFactory.UUID));

			assertThat(attachedFiles).hasSize(1).containsOnly(Map.entry(AttachmentType.ATTACHMENT, Set.of(expectedAttachment)));
		}

		@Test
		void shouldGetRepresentations() {
			var expectedAttachment = initAttachment(representation);

			var attachedFiles = getAttachedFiles(Set.of(FormcycleAttachmentTestFactory.UUID));

			assertThat(attachedFiles).hasSize(1).containsOnly(Map.entry(AttachmentType.ATTACHMENT, Set.of(expectedAttachment)));
		}

		@Test
		@DisplayName("should return all attachments as representations when no upload fields")
		@SneakyThrows
		void shouldGetAllRepresentations() {
			when(attachmentMapper.map(any())).thenAnswer(invocation -> mock(FormcycleAttachment.class));
			when(attachmentSupplier.get()).thenReturn(List.of(attachment, representation));

			var attachedFiles = getAttachedFiles(Collections.emptySet());

			assertThat(attachedFiles).hasSize(1);
			assertThat(attachedFiles.get(AttachmentType.REPRESENTATION)).hasSize(2);
		}

		@SneakyThrows
		private FormcycleAttachment initAttachment(Attachment attachment) {
			var expectedAttachment = FormcycleAttachmentTestFactory.create();
			when(attachmentMapper.map(any())).thenReturn(expectedAttachment);
			when(attachment.getUUID()).thenReturn(FormcycleAttachmentTestFactory.UUID);
			when(attachmentSupplier.get()).thenReturn(List.of(attachment));
			return expectedAttachment;
		}

		@SneakyThrows
		private Map<AttachmentType, Set<FormcycleAttachment>> getAttachedFiles(Set<String> ids) {
			return executor.getAttachedFiles(ids);
		}
	}

}